/*
  rios.c

  RIOS provides a simple, C-based approach to providing simple 
  multitasking  functionality  in  embedded  designs.
  
  Based on the article 
  "RIOS: A Lightweight Task Scheduler for Embedded Systems" 
  https://www.ics.uci.edu/~givargis/pubs/C50.pdf

  Initiator 17.04.2017 12:45:50 Detlev Tietjen

  List of Authors
  17.4.2017 V0.1 Detlev Tietjen, initial version
  22.3.2017 V0.2 Markus        , Addaption to timer1 for usage with Arduino Uno

*/

#include <stddef.h>
#include <stdbool.h>
#include <avr/interrupt.h>
#include "rios.h"

#define THREAD_DIVISOR ((F_CPU * THREAD_MS_PER_TICK + 500UL * 1024UL) / (1000UL * 1024UL))
#if (255 < THREAD_DIVISOR)
#error THREAD_MS_PER_TICK to high
#endif

/**
   structure to hold data of the thread control block (TCB). This structure is capsuled and should
   only accessed by setter and getter functions.
*/
struct threadTCBData_t {
  uint8_t     state;      ///< The last state of the thread
  bool      alreadyRunning; ///< Set to true as long the thread has not finished the last call
  threadFunc_t  func;     ///< Pointer to the thread function.
  void      *pArgument;   ///< Holds the argument of the thread function
  uint16_t    period;     ///< time between two calls in ticks
  uint16_t    elapsed;    ///< elapsed time since the last call in ticks
};

static uint8_t nextThreadIndex = 0;   ///<holds the number of defined threads
static struct threadTCBData_t thread[THREAD_MAX];///<array of thread control blocks (TCB)

pThread_t threadAdd(threadFunc_t f, void * pArgument, uint16_t periodTicks) {
  pThread_t rp = NULL;

  if ((0 < periodTicks) && (THREAD_MAX > nextThreadIndex)) {
    rp          = thread + nextThreadIndex++;
    rp->state       = 0;
    rp->alreadyRunning  = false;
    rp->func      = f;
    rp->pArgument     = pArgument;
    rp->period      = periodTicks;
    rp->elapsed     = periodTicks; //start initialization immediately
  }
  return rp;
}

void threadStart()
{
#ifdef USE_TIMER1
  OCR1A  = THREAD_DIVISOR;
  TCCR1A = 2; // TOP = OCR0A
  TCCR1B = 5; // prescaler = 1024
  TIMSK1 = (1 << OCIE1A); // enable timetick interrupt
#else
  OCR0A  = THREAD_DIVISOR;
  TCCR0A = 2; // TOP = OCR0A
  TCCR0B = 5; // prescaler = 1024
  TIMSK0 = (1 << OCIE0A); // enable timetick interrupt
#endif
  sei();
}

#ifdef USE_TIMER1
ISR(TIMER1_COMPA_vect, ISR_NOBLOCK)
#else
ISR(TIMER0_COMPA_vect, ISR_NOBLOCK)
#endif
{
  bool higherThreadAlreadyRunning = false;
  pThread_t pThread = thread;
  uint8_t threadCnt = nextThreadIndex; //total count of active tasks

  while (threadCnt--) {
    if (!higherThreadAlreadyRunning) {
      if (pThread->alreadyRunning) {
        higherThreadAlreadyRunning = true;
      } else {
        threadBeginCriticalBlock();
        if (pThread->elapsed >= pThread->period) {
          pThread->elapsed -= pThread->period;
          pThread->alreadyRunning = true;
        }
        threadEndCriticalBlock();

        if (pThread->alreadyRunning) // This could only be true if set in the block before
        {
          pThread->state = pThread->func(pThread->state,
                                         pThread->pArgument); //call the thread
          pThread->alreadyRunning = false;
        }
      }
    }
    threadBeginCriticalBlock();
    if (UINT16_MAX > pThread->elapsed)
      pThread->elapsed++;
    threadEndCriticalBlock();
    pThread++;
  }
}
